//-----------------------------------------------------------------------------
// Copyright (c) 2007-2008 dhpoware. All Rights Reserved.
//
// Permission is hereby granted, free of charge, to any person obtaining a
// copy of this software and associated documentation files (the "Software"),
// to deal in the Software without restriction, including without limitation
// the rights to use, copy, modify, merge, publish, distribute, sublicense,
// and/or sell copies of the Software, and to permit persons to whom the
// Software is furnished to do so, subject to the following conditions:
//
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
// OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
// FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
// IN THE SOFTWARE.
//-----------------------------------------------------------------------------
// 
// This is the fourth demo in the Direct3D camera demo series and it builds on
// the previous 2 demos in the series:
//
//  Direct3D Camera Demo - Part 2
//  (http://www.dhpoware.com/demos/d3dCamera2.html)
//
//  Direct3D Third Person Camera Demo - Part 1
//  (http://www.dhpoware.com/demos/d3dThirdPersonCamera1.html)
//
// This demo shows how various camera behaviors can be combined together. The
// following camera behaviors are implemented in this demo: first person,
// spectator, flight, and orbit. The first person camera behavior simulates
// the view from the perspective of the player. The spectator camera behavior
// simulates a floating camera that moves in the direction that the camera is
// looking. The flight camera behavior simulates the view from the cockpit of
// an airplane. Finally the orbit camera behavior simulates a third person view
// of the player and allows the camera to orbit the player.
//
// The camera class that accompanies this demo uses quaternions throughout. Of
// particular note is the class' implementation of the orbit camera behavior.
// The orbit camera implementation differs considerably from the third person
// camera model used in the Direct3D Third Person Camera Demo - Part 1.
//
// The Direct3D Third Person Camera Demo maintained an offset vector and used
// quaternions to rotate this offset vector when the camera was being rotated.
// The new camera position was then used to build a look-at style view matrix.
//
// This demo's orbit logic begins by setting the camera's orientation to that
// of the orbit target. Then rotating the camera will affect this orientation.
// So far this is exactly how the other camera behaviors (first person,
// spectator, and flight) work. But once the orbit camera is rotated the
// camera's view matrix is updated to accommodate the new orientation. The
// view matrix requires the camera position to perform a dot product with each
// of the three local axes. To implement a third person orbit style view of the
// target the camera's position is calculated here to be a fixed distance from
// the target's position. Specifically this demo takes the orbit target's world
// position and translates this point by m_orbitOffsetDistance world units
// along the camera's local negative Z axis.
//
//-----------------------------------------------------------------------------

#if !defined(WIN32_LEAN_AND_MEAN)
#define WIN32_LEAN_AND_MEAN
#endif

#if defined(_DEBUG)
#define D3D_DEBUG_INFO
#endif

#include <windows.h>
#include <d3d9.h>
#include <d3dx9.h>
#include <iomanip>
#include <sstream>
#include <stdexcept>
#include <string>

#if defined(_DEBUG)
#include <crtdbg.h>
#endif

#include "camera.h"
#include "input.h"
#include "mesh_utils.h"
#include "normal_mapping_utils.h"

//-----------------------------------------------------------------------------
// Macros.
//-----------------------------------------------------------------------------

#define SAFE_RELEASE(x) if (x != 0) { x->Release(); x = 0; }

//-----------------------------------------------------------------------------
// Constants.
//-----------------------------------------------------------------------------

#if !defined(CLEARTYPE_QUALITY)
#define CLEARTYPE_QUALITY 5
#endif

#define APP_TITLE   "D3D Camera Demo 3"

const float         FLOOR_WIDTH = 8.0f;
const float         FLOOR_HEIGHT = 8.0f;
const float         FLOOR_TILE_U = 8.0f;
const float         FLOOR_TILE_V = 8.0f;

const float         CAMERA_FOVX = 90.0f;
const float         CAMERA_OFFSET = 0.7f;
const float         CAMERA_ZFAR = 100.0f;
const float         CAMERA_ZNEAR = 0.1f;
const float         CAMERA_ZOOM_MIN = 1.5f;
const float         CAMERA_ZOOM_MAX = 5.0f;

const float         CAMERA_SPEED_FLIGHT_YAW = 100.0f;
const float         CAMERA_SPEED_ORBIT_ROLL = 100.0f;

const D3DXVECTOR3   CAMERA_ACCELERATION(8.0f, 8.0f, 8.0f);
const D3DXVECTOR3   CAMERA_VELOCITY(1.0f, 1.0f, 1.0f);

const D3DXVECTOR3   CAMERA_BOUNDS_MAX(FLOOR_WIDTH / 2.0f, 4.0f, FLOOR_HEIGHT / 2.0f);
const D3DXVECTOR3   CAMERA_BOUNDS_MIN(-FLOOR_WIDTH / 2.0f, CAMERA_OFFSET, -FLOOR_HEIGHT / 2.0f);

const float         LIGHT_RADIUS = max(FLOOR_WIDTH, FLOOR_HEIGHT);
const float         LIGHT_SPOT_INNER_CONE = D3DXToRadian(30.0f);
const float         LIGHT_SPOT_OUTER_CONE = D3DXToRadian(100.0f);
const D3DXVECTOR3   LIGHT_DIR(0.0f, -1.0f, 0.0f);
const D3DXVECTOR3   LIGHT_POS(0.0f, LIGHT_RADIUS * 0.5f, 0.0f);

const float         MOUSE_SPEED_X = 80.0f;
const float         MOUSE_SPEED_Y = 80.0f;

//-----------------------------------------------------------------------------
// Types.
//-----------------------------------------------------------------------------

struct Light
{
    float dir[3];
    float pos[3];
    float ambient[4];
    float diffuse[4];
    float specular[4];
    float spotInnerCone;
    float spotOuterCone;
    float radius;
};

struct Material
{
    float ambient[4];
    float diffuse[4];
    float emissive[4];
    float specular[4];
    float shininess;
};

//-----------------------------------------------------------------------------
// Globals.
//-----------------------------------------------------------------------------

HWND                         g_hWnd;
HINSTANCE                    g_hInstance;
IDirect3D9                  *g_pDirect3D;
IDirect3DDevice9            *g_pDevice;
ID3DXFont                   *g_pFont;
ID3DXEffect                 *g_pParallaxNormalMappingEffect;
ID3DXEffect                 *g_pPerPixelLightingEffect;
IDirect3DVertexDeclaration9 *g_pFloorVertexDeclaration;
IDirect3DVertexBuffer9      *g_pFloorVertexBuffer;
IDirect3DTexture9           *g_pFloorColorMapTexture;
IDirect3DTexture9           *g_pFloorNormalMapTexture;
IDirect3DTexture9           *g_pFloorHeightMapTexture;
IDirect3DTexture9           *g_pNullTexture;
DWORD                        g_msaaSamples;
DWORD                        g_maxAnisotrophy;
int                          g_framesPerSecond;
int                          g_windowWidth;
int                          g_windowHeight;
float                        g_globalAmbient[4] = {0.0f, 0.0f, 0.0f, 1.0f};
float                        g_parallaxScaleBias[2] = {0.04f, -0.03f};
bool                         g_enableParallax;
bool                         g_enableVerticalSync;
bool                         g_isFullScreen;
bool                         g_hasFocus;
bool                         g_displayHelp;
bool                         g_disableColorMapTexture;
bool                         g_disableSpecularFloor;
D3DPRESENT_PARAMETERS        g_params;
D3DCAPS9                     g_caps;
Camera                       g_camera;
NormalMappedQuad             g_floorQuad;
Mesh                         g_mesh;
D3DXQUATERNION               g_meshOrientation;
D3DXVECTOR3                  g_meshPosition;
D3DXVECTOR3                  g_velocity;

Light g_light =
{
    LIGHT_DIR.x, LIGHT_DIR.y, LIGHT_DIR.z,      // dir
    LIGHT_POS.x, LIGHT_POS.y, LIGHT_POS.z,      // pos
    1.0f, 1.0f, 1.0f, 1.0f,                     // ambient
    1.0f, 1.0f, 1.0f, 1.0f,                     // diffuse
    1.0f, 1.0f, 1.0f, 1.0f,                     // specular
    LIGHT_SPOT_INNER_CONE,                      // spotInnerCone
    LIGHT_SPOT_OUTER_CONE,                      // spotOuterCone
    LIGHT_RADIUS                                // radius
};

Material g_material =
{
    0.2f, 0.2f, 0.2f, 1.0f,                     // ambient
    0.8f, 0.8f, 0.8f, 1.0f,                     // diffuse
    0.0f, 0.0f, 0.0f, 1.0f,                     // emissive
    1.0f, 1.0f, 1.0f, 1.0f,                     // specular
    90.0f                                       // shininess
};

//-----------------------------------------------------------------------------
// Function Prototypes.
//-----------------------------------------------------------------------------

void    ChangeCameraBehavior(Camera::CameraBehavior behavior);
void    ChooseBestMSAAMode(D3DFORMAT backBufferFmt, D3DFORMAT depthStencilFmt,
                           BOOL windowed, D3DMULTISAMPLE_TYPE &type,
                           DWORD &qualityLevels, DWORD &samplesPerPixel);
void    Cleanup();
HWND    CreateAppWindow(const WNDCLASSEX &wcl, const char *pszTitle);
bool    CreateNullTexture(int width, int height, LPDIRECT3DTEXTURE9 &pTexture);
bool    DeviceIsValid();
float   GetElapsedTimeInSeconds();
void    GetMovementDirection(D3DXVECTOR3 &direction);
bool    Init();
void    InitApp();
void    InitCamera();
bool    InitD3D();
void    InitFloor();
void    InitFont();
bool    InitFont(const char *pszFont, int ptSize, LPD3DXFONT &pFont);
void    InitMesh();
void    InitShaders();
void    InitTextures();
bool    LoadShader(const char *pszFilename, LPD3DXEFFECT &pEffect);
void    Log(const char *pszMessage);
bool    MSAAModeSupported(D3DMULTISAMPLE_TYPE type, D3DFORMAT backBufferFmt,
                          D3DFORMAT depthStencilFmt, BOOL windowed,
                          DWORD &qualityLevels);
void    PerformCameraCollisionDetection();
void    ProcessUserInput();
void    RenderFloor();
void    RenderFrame();
void    RenderMesh();
void    RenderText();
bool    ResetDevice();
void    SetProcessorAffinity();
void    ToggleFullScreen();
void    UpdateCamera(float elapsedTimeSec);
void    UpdateFrame(float elapsedTimeSec);
void    UpdateFrameRate(float elapsedTimeSec);
void    UpdateParallaxNormalMappingEffectParameters();
void    UpdatePerPixelLightingEffectParameters();
LRESULT CALLBACK WindowProc(HWND hWnd, UINT msg, WPARAM wParam, LPARAM lParam);

//-----------------------------------------------------------------------------
// Functions.
//-----------------------------------------------------------------------------

int WINAPI WinMain(HINSTANCE hInstance, HINSTANCE hPrevInstance, LPSTR lpCmdLine, int nShowCmd)
{
#if defined _DEBUG
    _CrtSetDbgFlag(_CRTDBG_LEAK_CHECK_DF | _CRTDBG_ALLOC_MEM_DF);
    _CrtSetReportMode(_CRT_ASSERT, _CRTDBG_MODE_FILE);
    _CrtSetReportFile(_CRT_ASSERT, _CRTDBG_FILE_STDERR);
#endif

    MSG msg = {0};
    WNDCLASSEX wcl = {0};

    wcl.cbSize = sizeof(wcl);
    wcl.style = CS_OWNDC | CS_HREDRAW | CS_VREDRAW;
    wcl.lpfnWndProc = WindowProc;
    wcl.cbClsExtra = 0;
    wcl.cbWndExtra = 0;
    wcl.hInstance = g_hInstance = hInstance;
    wcl.hIcon = LoadIcon(0, IDI_APPLICATION);
    wcl.hCursor = LoadCursor(0, IDC_ARROW);
    wcl.hbrBackground = 0;
    wcl.lpszMenuName = 0;
    wcl.lpszClassName = "D3D9WindowClass";
    wcl.hIconSm = 0;

    if (!RegisterClassEx(&wcl))
        return 0;

    g_hWnd = CreateAppWindow(wcl, APP_TITLE);

    if (g_hWnd)
    {
        SetProcessorAffinity();

        if (Init())
        {
            ShowWindow(g_hWnd, nShowCmd);
            UpdateWindow(g_hWnd);

            while (true)
            {
                while (PeekMessage(&msg, 0, 0, 0, PM_REMOVE))
                {
                    if (msg.message == WM_QUIT)
                        break;

                    TranslateMessage(&msg);
                    DispatchMessage(&msg);
                }

                if (msg.message == WM_QUIT)
                    break;

                if (g_hasFocus)
                {
                    UpdateFrame(GetElapsedTimeInSeconds());

                    if (DeviceIsValid())
                        RenderFrame();
                }
                else
                {
                    WaitMessage();
                }
            }
        }

        Cleanup();
        UnregisterClass(wcl.lpszClassName, hInstance);
    }

    return static_cast<int>(msg.wParam);
}

LRESULT CALLBACK WindowProc(HWND hWnd, UINT msg, WPARAM wParam, LPARAM lParam)
{
    Keyboard::instance().handleMsg(hWnd, msg, wParam, lParam);

    switch (msg)
    {
    default:
        break;

    case WM_ACTIVATE:
        switch (wParam)
        {
        default:
            break;

        case WA_ACTIVE:
        case WA_CLICKACTIVE:
            g_hasFocus = true;
            break;

        case WA_INACTIVE:
            if (g_isFullScreen)
                ShowWindow(hWnd, SW_MINIMIZE);
            g_hasFocus = false;
            break;
        }
        break;

    case WM_DESTROY:
        PostQuitMessage(0);
        return 0;

    case WM_SIZE:
        g_windowWidth = static_cast<int>(LOWORD(lParam));
        g_windowHeight = static_cast<int>(HIWORD(lParam));
        break;
    }

    return DefWindowProc(hWnd, msg, wParam, lParam);
}

void ChangeCameraBehavior(Camera::CameraBehavior behavior)
{
    if (g_camera.getBehavior() == behavior)
        return;

    if (behavior == Camera::CAMERA_BEHAVIOR_ORBIT)
    {
        g_meshPosition = g_camera.getPosition();
        D3DXQuaternionInverse(&g_meshOrientation, &g_camera.getOrientation());
    }

    g_camera.setBehavior(behavior);

    if (behavior == Camera::CAMERA_BEHAVIOR_ORBIT)
        g_camera.rotate(0.0f, 30.0f, 0.0f);
}

void ChooseBestMSAAMode(D3DFORMAT backBufferFmt, D3DFORMAT depthStencilFmt,
                        BOOL windowed, D3DMULTISAMPLE_TYPE &type,
                        DWORD &qualityLevels, DWORD &samplesPerPixel)
{
    bool supported = false;

    struct MSAAMode
    {
        D3DMULTISAMPLE_TYPE type;
        DWORD samples;
    }
    multsamplingTypes[15] =
    {
        { D3DMULTISAMPLE_16_SAMPLES,  16 },
        { D3DMULTISAMPLE_15_SAMPLES,  15 },
        { D3DMULTISAMPLE_14_SAMPLES,  14 },
        { D3DMULTISAMPLE_13_SAMPLES,  13 },
        { D3DMULTISAMPLE_12_SAMPLES,  12 },
        { D3DMULTISAMPLE_11_SAMPLES,  11 },
        { D3DMULTISAMPLE_10_SAMPLES,  10 },
        { D3DMULTISAMPLE_9_SAMPLES,   9  },
        { D3DMULTISAMPLE_8_SAMPLES,   8  },
        { D3DMULTISAMPLE_7_SAMPLES,   7  },
        { D3DMULTISAMPLE_6_SAMPLES,   6  },
        { D3DMULTISAMPLE_5_SAMPLES,   5  },
        { D3DMULTISAMPLE_4_SAMPLES,   4  },
        { D3DMULTISAMPLE_3_SAMPLES,   3  },
        { D3DMULTISAMPLE_2_SAMPLES,   2  }
    };

    for (int i = 0; i < 15; ++i)
    {
        type = multsamplingTypes[i].type;

        supported = MSAAModeSupported(type, backBufferFmt, depthStencilFmt,
                        windowed, qualityLevels);

        if (supported)
        {
            samplesPerPixel = multsamplingTypes[i].samples;
            return;
        }
    }

    type = D3DMULTISAMPLE_NONE;
    qualityLevels = 0;
    samplesPerPixel = 1;
}

void Cleanup()
{
    g_mesh.destroy();
    SAFE_RELEASE(g_pNullTexture)
    SAFE_RELEASE(g_pFloorHeightMapTexture)
    SAFE_RELEASE(g_pFloorNormalMapTexture)
    SAFE_RELEASE(g_pFloorColorMapTexture)
    SAFE_RELEASE(g_pFloorVertexBuffer)
    SAFE_RELEASE(g_pFloorVertexDeclaration)
    SAFE_RELEASE(g_pParallaxNormalMappingEffect)
    SAFE_RELEASE(g_pPerPixelLightingEffect)
    SAFE_RELEASE(g_pFont)
    SAFE_RELEASE(g_pDevice)
    SAFE_RELEASE(g_pDirect3D)
}

HWND CreateAppWindow(const WNDCLASSEX &wcl, const char *pszTitle)
{
    // Create a window that is centered on the desktop. It's exactly 1/4 the
    // size of the desktop. Don't allow it to be resized.

    DWORD wndExStyle = WS_EX_OVERLAPPEDWINDOW;
    DWORD wndStyle = WS_OVERLAPPED | WS_CAPTION | WS_SYSMENU |
                        WS_MINIMIZEBOX | WS_CLIPCHILDREN | WS_CLIPSIBLINGS;

    HWND hWnd = CreateWindowEx(wndExStyle, wcl.lpszClassName, pszTitle,
                    wndStyle, 0, 0, 0, 0, 0, 0, wcl.hInstance, 0);

    if (hWnd)
    {
        int screenWidth = GetSystemMetrics(SM_CXSCREEN);
        int screenHeight = GetSystemMetrics(SM_CYSCREEN);
        int halfScreenWidth = screenWidth / 2;
        int halfScreenHeight = screenHeight / 2;
        int left = (screenWidth - halfScreenWidth) / 2;
        int top = (screenHeight - halfScreenHeight) / 2;
        RECT rc = {0};

        SetRect(&rc, left, top, left + halfScreenWidth, top + halfScreenHeight);
        AdjustWindowRectEx(&rc, wndStyle, FALSE, wndExStyle);
        MoveWindow(hWnd, rc.left, rc.top, rc.right - rc.left, rc.bottom - rc.top, TRUE);

        GetClientRect(hWnd, &rc);
        g_windowWidth = rc.right - rc.left;
        g_windowHeight = rc.bottom - rc.top;
    }

    return hWnd;
}

bool CreateNullTexture(int width, int height, LPDIRECT3DTEXTURE9 &pTexture)
{
    // Create an empty white texture. This texture is applied to geometry
    // that doesn't have any texture maps. This trick allows the same shader to
    // be used to draw the geometry with and without textures applied.

    HRESULT hr = D3DXCreateTexture(g_pDevice, width, height, 0, 0,
        D3DFMT_X8R8G8B8, D3DPOOL_MANAGED, &pTexture);

    if (FAILED(hr))
        return false;

    LPDIRECT3DSURFACE9 pSurface = 0;

    if (SUCCEEDED(pTexture->GetSurfaceLevel(0, &pSurface)))
    {
        D3DLOCKED_RECT rcLock = {0};

        if (SUCCEEDED(pSurface->LockRect(&rcLock, 0, 0)))
        {
            BYTE *pPixels = static_cast<BYTE*>(rcLock.pBits);
            int widthInBytes = width * 4;

            if (widthInBytes == rcLock.Pitch)
            {
                memset(pPixels, 0xff, widthInBytes * height);
            }
            else
            {
                for (int y = 0; y < height; ++y)
                    memset(&pPixels[y * rcLock.Pitch], 0xff, rcLock.Pitch);
            }

            pSurface->UnlockRect();
            pSurface->Release();
            return true;
        }

        pSurface->Release();
    }

    pTexture->Release();
    return false;
}

bool DeviceIsValid()
{
    HRESULT hr = g_pDevice->TestCooperativeLevel();

    if (FAILED(hr))
    {
        if (hr == D3DERR_DEVICENOTRESET)
            return ResetDevice();
    }

    return true;
}

float GetElapsedTimeInSeconds()
{
    // Returns the elapsed time (in seconds) since the last time this function
    // was called. This elaborate setup is to guard against large spikes in
    // the time returned by QueryPerformanceCounter().

    static const int MAX_SAMPLE_COUNT = 50;

    static float frameTimes[MAX_SAMPLE_COUNT];
    static float timeScale = 0.0f;
    static float actualElapsedTimeSec = 0.0f;
    static INT64 freq = 0;
    static INT64 lastTime = 0;
    static int sampleCount = 0;
    static bool initialized = false;

    INT64 time = 0;
    float elapsedTimeSec = 0.0f;

    if (!initialized)
    {
        initialized = true;
        QueryPerformanceFrequency(reinterpret_cast<LARGE_INTEGER*>(&freq));
        QueryPerformanceCounter(reinterpret_cast<LARGE_INTEGER*>(&lastTime));
        timeScale = 1.0f / freq;
    }

    QueryPerformanceCounter(reinterpret_cast<LARGE_INTEGER*>(&time));
    elapsedTimeSec = (time - lastTime) * timeScale;
    lastTime = time;

    if (fabsf(elapsedTimeSec - actualElapsedTimeSec) < 1.0f)
    {
        memmove(&frameTimes[1], frameTimes, sizeof(frameTimes) - sizeof(frameTimes[0]));
        frameTimes[0] = elapsedTimeSec;

        if (sampleCount < MAX_SAMPLE_COUNT)
            ++sampleCount;
    }

    actualElapsedTimeSec = 0.0f;

    for (int i = 0; i < sampleCount; ++i)
        actualElapsedTimeSec += frameTimes[i];

    if (sampleCount > 0)
        actualElapsedTimeSec /= sampleCount;

    return actualElapsedTimeSec;
}

void GetMovementDirection(D3DXVECTOR3 &direction)
{
    static bool moveForwardsPressed = false;
    static bool moveBackwardsPressed = false;
    static bool moveRightPressed = false;
    static bool moveLeftPressed = false;
    static bool moveUpPressed = false;
    static bool moveDownPressed = false;

    D3DXVECTOR3 velocity = g_camera.getCurrentVelocity();
    Keyboard &keyboard = Keyboard::instance();

    direction.x = direction.y = direction.z = 0.0f;

    if (keyboard.keyDown(Keyboard::KEY_UP) || keyboard.keyDown(Keyboard::KEY_W))
    {
        if (!moveForwardsPressed)
        {
            moveForwardsPressed = true;
            g_camera.setCurrentVelocity(velocity.x, velocity.y, 0.0f);
        }

        direction.z += 1.0f;
    }
    else
    {
        moveForwardsPressed = false;
    }

    if (keyboard.keyDown(Keyboard::KEY_DOWN) || keyboard.keyDown(Keyboard::KEY_S))
    {
        if (!moveBackwardsPressed)
        {
            moveBackwardsPressed = true;
            g_camera.setCurrentVelocity(velocity.x, velocity.y, 0.0f);
        }

        direction.z -= 1.0f;
    }
    else
    {
        moveBackwardsPressed = false;
    }

    if (keyboard.keyDown(Keyboard::KEY_RIGHT) || keyboard.keyDown(Keyboard::KEY_D))
    {
        if (!moveRightPressed)
        {
            moveRightPressed = true;
            g_camera.setCurrentVelocity(0.0f, velocity.y, velocity.z);
        }

        direction.x += 1.0f;
    }
    else
    {
        moveRightPressed = false;
    }

    if (keyboard.keyDown(Keyboard::KEY_LEFT) || keyboard.keyDown(Keyboard::KEY_A))
    {
        if (!moveLeftPressed)
        {
            moveLeftPressed = true;
            g_camera.setCurrentVelocity(0.0f, velocity.y, velocity.z);
        }

        direction.x -= 1.0f;
    }
    else
    {
        moveLeftPressed = false;
    }

    if (keyboard.keyDown(Keyboard::KEY_E) || keyboard.keyDown(Keyboard::KEY_PAGEUP))
    {
        if (!moveUpPressed)
        {
            moveUpPressed = true;
            g_camera.setCurrentVelocity(velocity.x, 0.0f, velocity.z);
        }

        direction.y += 1.0f;
    }
    else
    {
        moveUpPressed = false;
    }

    if (keyboard.keyDown(Keyboard::KEY_Q) || keyboard.keyDown(Keyboard::KEY_PAGEDOWN))
    {
        if (!moveDownPressed)
        {
            moveDownPressed = true;
            g_camera.setCurrentVelocity(velocity.x, 0.0f, velocity.z);
        }

        direction.y -= 1.0f;
    }
    else
    {
        moveDownPressed = false;
    }
}

bool Init()
{
    if (!InitD3D())
    {
        Log("Direct3D initialization failed!");
        return false;
    }

    try
    {
        InitApp();
    }
    catch (const std::exception &e)
    {
        std::ostringstream msg;

        msg << "Application initialization failed!" << std::endl << std::endl;
        msg << e.what();

        Log(msg.str().c_str());
        return false;
    }

    return true;
}

void InitApp()
{
    InitFont();
    InitMesh();
    InitCamera();
    InitFloor();
    InitShaders();
    InitTextures();
}

void InitCamera()
{
    g_camera.perspective(CAMERA_FOVX,
        static_cast<float>(g_windowWidth) / static_cast<float>(g_windowHeight),
        CAMERA_ZNEAR, CAMERA_ZFAR);

    g_camera.setPosition(D3DXVECTOR3(0.0f, CAMERA_OFFSET, 0.0f));
    g_camera.setOrbitMinZoom(CAMERA_ZOOM_MIN);
    g_camera.setOrbitMaxZoom(CAMERA_ZOOM_MAX);
    g_camera.setOrbitOffsetDistance(CAMERA_ZOOM_MIN + (CAMERA_ZOOM_MAX - CAMERA_ZOOM_MIN) * 0.3f);

    g_camera.setAcceleration(CAMERA_ACCELERATION);
    g_camera.setVelocity(CAMERA_VELOCITY);

    ChangeCameraBehavior(Camera::CAMERA_BEHAVIOR_ORBIT);
}

bool InitD3D()
{
    HRESULT hr = 0;
    D3DDISPLAYMODE desktop = {0};

    g_pDirect3D = Direct3DCreate9(D3D_SDK_VERSION);

    if (!g_pDirect3D)
        return false;

    // Just use the current desktop display mode.
    hr = g_pDirect3D->GetAdapterDisplayMode(D3DADAPTER_DEFAULT, &desktop);

    if (FAILED(hr))
    {
        g_pDirect3D->Release();
        g_pDirect3D = 0;
        return false;
    }

    // Setup Direct3D for windowed rendering.
    g_params.BackBufferWidth = 0;
    g_params.BackBufferHeight = 0;
    g_params.BackBufferFormat = desktop.Format;
    g_params.BackBufferCount = 1;
    g_params.hDeviceWindow = g_hWnd;
    g_params.Windowed = TRUE;
    g_params.EnableAutoDepthStencil = TRUE;
    g_params.AutoDepthStencilFormat = D3DFMT_D24S8;
    g_params.Flags = D3DPRESENTFLAG_DISCARD_DEPTHSTENCIL;
    g_params.FullScreen_RefreshRateInHz = 0;

    if (g_enableVerticalSync)
        g_params.PresentationInterval = D3DPRESENT_INTERVAL_DEFAULT;
    else
        g_params.PresentationInterval = D3DPRESENT_INTERVAL_IMMEDIATE;

    // Swap effect must be D3DSWAPEFFECT_DISCARD for multi-sampling support.
    g_params.SwapEffect = D3DSWAPEFFECT_DISCARD;

    // Select the highest quality multi-sample anti-aliasing (MSAA) mode.
    ChooseBestMSAAMode(g_params.BackBufferFormat, g_params.AutoDepthStencilFormat,
        g_params.Windowed, g_params.MultiSampleType, g_params.MultiSampleQuality,
        g_msaaSamples);

    // Most modern video cards should have no problems creating pure devices.
    // Note that by creating a pure device we lose the ability to debug vertex
    // and pixel shaders.
    hr = g_pDirect3D->CreateDevice(D3DADAPTER_DEFAULT, D3DDEVTYPE_HAL, g_hWnd,
        D3DCREATE_HARDWARE_VERTEXPROCESSING | D3DCREATE_PUREDEVICE,
        &g_params, &g_pDevice);

    if (FAILED(hr))
    {
        // Fall back to software vertex processing for less capable hardware.
        // Note that in order to debug vertex shaders we must use a software
        // vertex processing device.
        hr = g_pDirect3D->CreateDevice(D3DADAPTER_DEFAULT, D3DDEVTYPE_HAL,
            g_hWnd, D3DCREATE_SOFTWARE_VERTEXPROCESSING, &g_params, &g_pDevice);
    }

    if (FAILED(hr))
    {
        g_pDirect3D->Release();
        g_pDirect3D = 0;
        return false;
    }

    // Prefer anisotropic texture filtering if it's supported.
    if (SUCCEEDED(g_pDevice->GetDeviceCaps(&g_caps)))
    {
        if (g_caps.RasterCaps & D3DPRASTERCAPS_ANISOTROPY)
            g_maxAnisotrophy = g_caps.MaxAnisotropy;
        else
            g_maxAnisotrophy = 1;
    }

    return true;
}

void InitFloor()
{
    HRESULT hr = 0;
    NormalMappedQuad::Vertex *pVertices = 0;

    g_floorQuad.generate(D3DXVECTOR3(0.0f, 0.0f, 0.0f),
        D3DXVECTOR3(0.0f, 1.0f, 0.0f), D3DXVECTOR3(0.0f, 0.0f, 1.0f),
        FLOOR_WIDTH, FLOOR_HEIGHT, FLOOR_TILE_U, FLOOR_TILE_V);

    hr = g_pDevice->CreateVertexDeclaration(g_floorQuad.getVertexElements(),
        &g_pFloorVertexDeclaration);

    if (FAILED(hr))
        throw std::runtime_error("Failed to create floor vertex declaration.");

    int totalBytes = g_floorQuad.getVertexSize() * g_floorQuad.getVertexCount();

    hr = g_pDevice->CreateVertexBuffer(totalBytes, 0, 0,
        D3DPOOL_MANAGED, &g_pFloorVertexBuffer, 0);

    if (FAILED(hr))
        throw std::runtime_error("Failed to create floor vertex buffer.");

    hr = g_pFloorVertexBuffer->Lock(0, 0, reinterpret_cast<void**>(&pVertices), 0);

    if (FAILED(hr))
        throw std::runtime_error("Failed to lock floor vertex buffer.");

    memcpy(pVertices, g_floorQuad.getVertices(), totalBytes);
    g_pFloorVertexBuffer->Unlock();
}

void InitFont()
{
    if (!InitFont("Arial", 8, g_pFont))
        throw std::runtime_error("Failed to create font.");
}

bool InitFont(const char *pszFont, int ptSize, LPD3DXFONT &pFont)
{
    static DWORD dwQuality = 0;

    // Prefer ClearType font quality if available.

    if (!dwQuality)
    {
        DWORD dwVersion = GetVersion();
        DWORD dwMajorVersion = static_cast<DWORD>((LOBYTE(LOWORD(dwVersion))));
        DWORD dwMinorVersion = static_cast<DWORD>((HIBYTE(LOWORD(dwVersion))));

        // Windows XP and higher will support ClearType quality fonts.
        if (dwMajorVersion >= 6 || (dwMajorVersion == 5 && dwMinorVersion == 1))
            dwQuality = CLEARTYPE_QUALITY;
        else
            dwQuality = ANTIALIASED_QUALITY;
    }

    int logPixelsY = 0;

    // Convert from font point size to pixel size.

    if (HDC hDC = GetDC((0)))
    {
        logPixelsY = GetDeviceCaps(hDC, LOGPIXELSY);
        ReleaseDC(0, hDC);
    }

    int fontCharHeight = -ptSize * logPixelsY / 72;

    // Now create the font. Prefer anti-aliased text.

    HRESULT hr = D3DXCreateFont(
        g_pDevice,
        fontCharHeight,                 // height
        0,                              // width
        FW_BOLD,                        // weight
        1,                              // mipmap levels
        FALSE,                          // italic
        DEFAULT_CHARSET,                // char set
        OUT_DEFAULT_PRECIS,             // output precision
        dwQuality,                      // quality
        DEFAULT_PITCH | FF_DONTCARE,    // pitch and family
        pszFont,                        // font name
        &pFont);

    return SUCCEEDED(hr) ? true : false;
}

void InitMesh()
{
    if (!g_mesh.create("Content/Models/bigship1.x", g_pDevice))
        throw std::runtime_error("Failed to load bigship1.x.");

    if (!g_mesh.normalize())
        throw std::runtime_error("Failed to normalize bigship1.x.");

    if (!g_mesh.rotate(180.0f, 0.0f, 1.0f, 0.0f))
        throw std::runtime_error("Failed to rotate bigship1.x 180 degrees about Y axis.");
}

void InitShaders()
{
    if (!LoadShader("Content/Shaders/parallax_normal_mapping.fx", g_pParallaxNormalMappingEffect))
        throw std::runtime_error("Failed to load parallax_normal_mapping.fx.");

    if (!LoadShader("Content/Shaders/blinn_phong.fx", g_pPerPixelLightingEffect))
        throw std::runtime_error("Failed to load blinn_phong.fx.");
}

void InitTextures()
{
    HRESULT hr;

    hr = D3DXCreateTextureFromFile(g_pDevice,
        "Content/Textures/floor_color_map.dds", &g_pFloorColorMapTexture);

    if (FAILED(hr))
        throw std::runtime_error("Failed to load floor_color_map.dds.");

    hr = D3DXCreateTextureFromFile(g_pDevice,
        "Content/Textures/floor_height_map.dds", &g_pFloorHeightMapTexture);

    if (FAILED(hr))
        throw std::runtime_error("Failed to load floor_height_map.dds.");

    hr = D3DXCreateTextureFromFile(g_pDevice,
        "Content/Textures/floor_normal_map.dds", &g_pFloorNormalMapTexture);

    if (FAILED(hr))
        throw std::runtime_error("Failed to load floor_normal_map.dds.");

    if (!CreateNullTexture(2, 2, g_pNullTexture))
        throw std::runtime_error("Failed to create null texture.");
}

bool LoadShader(const char *pszFilename, LPD3DXEFFECT &pEffect)
{
    ID3DXBuffer *pCompilationErrors = 0;
    DWORD dwShaderFlags = D3DXFX_NOT_CLONEABLE | D3DXSHADER_NO_PRESHADER;

    // Both vertex and pixel shaders can be debugged. To enable shader
    // debugging add the following flag to the dwShaderFlags variable:
    //      dwShaderFlags |= D3DXSHADER_DEBUG;
    //
    // Vertex shaders can be debugged with either the REF device or a device
    // created for software vertex processing (i.e., the IDirect3DDevice9
    // object must be created with the D3DCREATE_SOFTWARE_VERTEXPROCESSING
    // behavior). Pixel shaders can be debugged only using the REF device.
    //
    // To enable vertex shader debugging add the following flag to the
    // dwShaderFlags variable:
    //     dwShaderFlags |= D3DXSHADER_FORCE_VS_SOFTWARE_NOOPT;
    //
    // To enable pixel shader debugging add the following flag to the
    // dwShaderFlags variable:
    //     dwShaderFlags |= D3DXSHADER_FORCE_PS_SOFTWARE_NOOPT;

    HRESULT hr = D3DXCreateEffectFromFile(g_pDevice, pszFilename, 0, 0,
                    dwShaderFlags, 0, &pEffect, &pCompilationErrors);

    if (FAILED(hr))
    {
        if (pCompilationErrors)
        {
            std::string compilationErrors(static_cast<const char *>(
                pCompilationErrors->GetBufferPointer()));

            pCompilationErrors->Release();
            throw std::runtime_error(compilationErrors);
        }
    }

    if (pCompilationErrors)
        pCompilationErrors->Release();

    return pEffect != 0;
}

void Log(const char *pszMessage)
{
    MessageBox(0, pszMessage, "Error", MB_ICONSTOP);
}

bool MSAAModeSupported(D3DMULTISAMPLE_TYPE type, D3DFORMAT backBufferFmt,
                       D3DFORMAT depthStencilFmt, BOOL windowed,
                       DWORD &qualityLevels)
{
    DWORD backBufferQualityLevels = 0;
    DWORD depthStencilQualityLevels = 0;

    HRESULT hr = g_pDirect3D->CheckDeviceMultiSampleType(D3DADAPTER_DEFAULT,
        D3DDEVTYPE_HAL, backBufferFmt, windowed, type,
        &backBufferQualityLevels);

    if (SUCCEEDED(hr))
    {
        hr = g_pDirect3D->CheckDeviceMultiSampleType(D3DADAPTER_DEFAULT,
            D3DDEVTYPE_HAL, depthStencilFmt, windowed, type,
            &depthStencilQualityLevels);

        if (SUCCEEDED(hr))
        {
            if (backBufferQualityLevels == depthStencilQualityLevels)
            {
                // The valid range is between zero and one less than the level
                // returned by IDirect3D9::CheckDeviceMultiSampleType().

                if (backBufferQualityLevels > 0)
                    qualityLevels = backBufferQualityLevels - 1;
                else
                    qualityLevels = backBufferQualityLevels;

                return true;
            }
        }
    }

    return false;
}

void PerformCameraCollisionDetection()
{
    // Very simple collision detection to prevent the first person, spectator
    // and flight cameras from moving too far outside of the scene. We don't
    // worry about the orbiting camera here because the min and max zoom
    // constants already limits how far the orbiting camera can zoom in/out
    // of the scene.

    if (g_camera.getBehavior() != Camera::CAMERA_BEHAVIOR_ORBIT)
    {
        const D3DXVECTOR3 &pos = g_camera.getPosition();
        D3DXVECTOR3 newPos(pos);

        if (pos.x > CAMERA_BOUNDS_MAX.x)
            newPos.x = CAMERA_BOUNDS_MAX.x;

        if (pos.x < CAMERA_BOUNDS_MIN.x)
            newPos.x = CAMERA_BOUNDS_MIN.x;

        if (pos.y > CAMERA_BOUNDS_MAX.y)
            newPos.y = CAMERA_BOUNDS_MAX.y;

        if (pos.y < CAMERA_BOUNDS_MIN.y)
            newPos.y = CAMERA_BOUNDS_MIN.y;

        if (pos.z > CAMERA_BOUNDS_MAX.z)
            newPos.z = CAMERA_BOUNDS_MAX.z;

        if (pos.z < CAMERA_BOUNDS_MIN.z)
            newPos.z = CAMERA_BOUNDS_MIN.z;

        g_camera.setPosition(newPos);
    }
}

void ProcessUserInput()
{
    Keyboard &keyboard = Keyboard::instance();
    Mouse &mouse = Mouse::instance();

    if (keyboard.keyPressed(Keyboard::KEY_ESCAPE))
        PostMessage(g_hWnd, WM_CLOSE, 0, 0);

    if (keyboard.keyPressed(Keyboard::KEY_BACKSPACE))
    {
        switch (g_camera.getBehavior())
        {
        default:
            break;

        case Camera::CAMERA_BEHAVIOR_FLIGHT:
            g_camera.undoRoll();
            break;

        case Camera::CAMERA_BEHAVIOR_ORBIT:
            if (!g_camera.preferTargetYAxisOrbiting())
                g_camera.undoRoll();
            break;
        }
    }

    if (keyboard.keyPressed(Keyboard::KEY_SPACE))
    {
        if (g_camera.getBehavior() == Camera::CAMERA_BEHAVIOR_ORBIT)
            g_camera.setPreferTargetYAxisOrbiting(!g_camera.preferTargetYAxisOrbiting());
    }

    if (keyboard.keyPressed(Keyboard::KEY_1))
        ChangeCameraBehavior(Camera::CAMERA_BEHAVIOR_FIRST_PERSON);

    if (keyboard.keyPressed(Keyboard::KEY_2))
        ChangeCameraBehavior(Camera::CAMERA_BEHAVIOR_SPECTATOR);

    if (keyboard.keyPressed(Keyboard::KEY_3))
        ChangeCameraBehavior(Camera::CAMERA_BEHAVIOR_FLIGHT);

    if (keyboard.keyPressed(Keyboard::KEY_4))
        ChangeCameraBehavior(Camera::CAMERA_BEHAVIOR_ORBIT);

    if (keyboard.keyPressed(Keyboard::KEY_H))
        g_displayHelp = !g_displayHelp;

    if (keyboard.keyPressed(Keyboard::KEY_P))
        g_enableParallax = !g_enableParallax;

    if (keyboard.keyPressed(Keyboard::KEY_T))
        g_disableColorMapTexture = !g_disableColorMapTexture;

    if (keyboard.keyPressed(Keyboard::KEY_ADD) || keyboard.keyPressed(Keyboard::KEY_NUMPAD_ADD))
    {
        g_camera.setRotationSpeed(g_camera.getRotationSpeed() + 0.01f);

        if (g_camera.getRotationSpeed() > 1.0f)
            g_camera.setRotationSpeed(1.0f);
    }     

    if (keyboard.keyPressed(Keyboard::KEY_MINUS) || keyboard.keyPressed(Keyboard::KEY_NUMPAD_MINUS))
    {
        g_camera.setRotationSpeed(g_camera.getRotationSpeed() - 0.01f);

        if (g_camera.getRotationSpeed() <= 0.0f)
            g_camera.setRotationSpeed(0.01f);
    }

    if (keyboard.keyPressed(Keyboard::KEY_PERIOD))
    {
        mouse.setWeightModifier(mouse.weightModifier() + 0.1f);

        if (mouse.weightModifier() > 1.0f)
            mouse.setWeightModifier(1.0f);
    }     

    if (keyboard.keyPressed(Keyboard::KEY_COMMA))
    {
        mouse.setWeightModifier(mouse.weightModifier() - 0.1f);

        if (mouse.weightModifier() < 0.0f)
            mouse.setWeightModifier(0.0f);
    }

    if (keyboard.keyPressed(Keyboard::KEY_M))
        mouse.smoothMouse(!mouse.isMouseSmoothing());

    if (keyboard.keyDown(Keyboard::KEY_LALT) || keyboard.keyDown(Keyboard::KEY_RALT))
    {
        if (keyboard.keyPressed(Keyboard::KEY_ENTER))
            ToggleFullScreen();
    }
}

void RenderFloor()
{
    D3DXHANDLE hTechnique;

    if (g_enableParallax)
        hTechnique = g_pParallaxNormalMappingEffect->GetTechniqueByName("ParallaxNormalMappingSpotLighting");
    else
        hTechnique = g_pParallaxNormalMappingEffect->GetTechniqueByName("NormalMappingSpotLighting");

    if (FAILED(g_pParallaxNormalMappingEffect->SetTechnique(hTechnique)))
        return;

    g_pDevice->SetVertexDeclaration(g_pFloorVertexDeclaration);
    g_pDevice->SetStreamSource(0, g_pFloorVertexBuffer, 0, g_floorQuad.getVertexSize());

    g_pDevice->SetTexture(0, g_pFloorColorMapTexture);
    g_pDevice->SetTexture(1, g_pFloorNormalMapTexture);

    if (g_enableParallax)
        g_pDevice->SetTexture(2, g_pFloorHeightMapTexture);

    UINT totalPasses = 0;

    if (FAILED(g_pParallaxNormalMappingEffect->Begin(&totalPasses, 0)))
        return;

    for (UINT i = 0; i < totalPasses; ++i)
    {
        if (SUCCEEDED(g_pParallaxNormalMappingEffect->BeginPass(i)))
        {
            g_pDevice->DrawPrimitive(D3DPT_TRIANGLELIST, 0, g_floorQuad.getPrimitiveCount());
            g_pParallaxNormalMappingEffect->EndPass();
        }
    }

    g_pParallaxNormalMappingEffect->End();
}

void RenderFrame()
{
    g_pDevice->Clear(0, 0, D3DCLEAR_TARGET | D3DCLEAR_ZBUFFER, 0, 1.0f, 0);

    if (FAILED(g_pDevice->BeginScene()))
        return;

    RenderFloor();

    if (g_camera.getBehavior() == Camera::CAMERA_BEHAVIOR_ORBIT)
        RenderMesh();

    RenderText();

    g_pDevice->EndScene();
    g_pDevice->Present(0, 0, 0, 0);
}

void RenderMesh()
{
    if (FAILED(g_pPerPixelLightingEffect->SetTechnique("PerPixelDirectionalLighting")))
        return;

    UINT totalPasses = 0;

    if (FAILED(g_pPerPixelLightingEffect->Begin(&totalPasses, 0)))
        return;

    for (UINT i = 0; i < totalPasses; ++i)
    {
        if (FAILED(g_pPerPixelLightingEffect->BeginPass(i)))
            break;

        for (int j = 0; j < static_cast<int>(g_mesh.materials.size()); ++j)
        {
            g_pPerPixelLightingEffect->SetValue("material.ambient",
                &g_mesh.materials[j].Ambient, sizeof(g_mesh.materials[j].Ambient));
            g_pPerPixelLightingEffect->SetValue("material.diffuse",
                &g_mesh.materials[j].Diffuse, sizeof(g_mesh.materials[j].Diffuse));
            g_pPerPixelLightingEffect->SetValue("material.emissive",
                &g_mesh.materials[j].Emissive, sizeof(g_mesh.materials[j].Emissive));
            g_pPerPixelLightingEffect->SetValue("material.specular",
                &g_mesh.materials[j].Specular, sizeof(g_mesh.materials[j].Specular));
            g_pPerPixelLightingEffect->SetFloat("material.shininess",
                g_mesh.materials[j].Power);

            g_pPerPixelLightingEffect->CommitChanges();
            g_mesh.pMesh->DrawSubset(j);
        }

        g_pPerPixelLightingEffect->EndPass();
    }

    g_pPerPixelLightingEffect->End();
}

void RenderText()
{
    std::ostringstream output;
    RECT rcClient;

    if (g_displayHelp)
    {
        output
            << "Press 1 to switch to first person behavior" << std::endl
            << "Press 2 to switch to spectator behavior" << std::endl
            << "Press 3 to switch to flight behavior" << std::endl
            << "Press 4 to switch to orbit behavior" << std::endl
            << std::endl
            << "First Person and Spectator behaviors" << std::endl
            << "  Press W and S to move forwards and backwards" << std::endl
            << "  Press A and D to strafe left and right" << std::endl
            << "  Press E and Q to move up and down" << std::endl
            << "  Move mouse to free look" << std::endl
            << std::endl
            << "Flight behavior" << std::endl
            << "  Press W and S to move forwards and backwards" << std::endl
            << "  Press A and D to yaw left and right" << std::endl
            << "  Press E and Q to move up and down" << std::endl
            << "  Move mouse up and down to change pitch" << std::endl
            << "  Move mouse left and right to change roll" << std::endl
            << std::endl
            << "Orbit behavior" << std::endl
            << "  Press SPACE to enable/disable target Y axis orbiting" << std::endl
            << "  Move mouse to orbit the model" << std::endl
            << "  Mouse wheel to zoom in and out" << std::endl
            << std::endl
            << "Press M to enable/disable mouse smoothing" << std::endl
            << "Press P to toggle between parallax normal mapping and normal mapping" << std::endl
            << "Press T to enable/disable the floor color map texture" << std::endl
            << "Press + and - to change camera rotation speed" << std::endl
            << "Press , and . to change mouse sensitivity" << std::endl
            << "Press SPACE to toggle between target y-axis and free orbiting" << std::endl
            << "Press BACKSPACE or middle mouse button to level camera" << std::endl
            << "Press ALT and ENTER to toggle full screen" << std::endl
            << "Press ESC to exit" << std::endl
            << std::endl
            << "Press H to hide help";
    }
    else
    {
        const char *pszCurrentBehavior = 0;
        const char *pszFloorEffect = 0;
        const char *pszOrbitStyle = 0;
        const Mouse &mouse = Mouse::instance();

        switch (g_camera.getBehavior())
        {
        default:
            pszCurrentBehavior = "Unknown";
            break;

        case Camera::CAMERA_BEHAVIOR_FIRST_PERSON:
            pszCurrentBehavior = "First Person";
            break;

        case Camera::CAMERA_BEHAVIOR_SPECTATOR:
            pszCurrentBehavior = "Spectator";
            break;

        case Camera::CAMERA_BEHAVIOR_FLIGHT:
            pszCurrentBehavior = "Flight";
            break;

        case Camera::CAMERA_BEHAVIOR_ORBIT:
            pszCurrentBehavior = "Orbit";
            break;
        }

        if (g_enableParallax)
            pszFloorEffect = "Parallax normal mapping";
        else
            pszFloorEffect = "Normal mapping";

        if (g_camera.preferTargetYAxisOrbiting())
            pszOrbitStyle = "Target Y axis";
        else
            pszOrbitStyle = "Free";

        output.setf(std::ios::fixed, std::ios::floatfield);
        output << std::setprecision(2);

        output
            << "FPS: " << g_framesPerSecond << std::endl
            << "Multisample anti-aliasing: " << g_msaaSamples << "x" << std::endl
            << "Anisotropic filtering: " << g_maxAnisotrophy << "x" << std::endl
            << "Floor effect: " << pszFloorEffect << std::endl
            << std::endl
            << "Camera" << std::endl
            << "  Position:"
            << " x:" << g_camera.getPosition().x
            << " y:" << g_camera.getPosition().y
            << " z:" << g_camera.getPosition().z << std::endl
            << "  Velocity:"
            << " x:" << g_camera.getCurrentVelocity().x
            << " y:" << g_camera.getCurrentVelocity().y
            << " z:" << g_camera.getCurrentVelocity().z << std::endl
            << "  Behavior: " << pszCurrentBehavior << std::endl
            << "  Rotation speed: " << g_camera.getRotationSpeed() << std::endl
            << "  Orbit style: " << pszOrbitStyle << std::endl
            << std::endl
            << "Mouse" << std::endl
            << "  Smoothing: " << (mouse.isMouseSmoothing() ? "enabled" : "disabled") << std::endl
            << "  Sensitivity: " << mouse.weightModifier() << std::endl
            << std::endl
            << "Press H to display help";
    }    

    GetClientRect(g_hWnd, &rcClient);
    rcClient.left += 4;
    rcClient.top += 2;

    g_pFont->DrawText(0, output.str().c_str(), -1, &rcClient,
        DT_EXPANDTABS | DT_LEFT, D3DCOLOR_XRGB(255, 255, 0));
}

bool ResetDevice()
{
    if (FAILED(g_pParallaxNormalMappingEffect->OnLostDevice()))
        return false;

    if (FAILED(g_pPerPixelLightingEffect->OnLostDevice()))
        return false;

    if (FAILED(g_pFont->OnLostDevice()))
        return false;

    if (FAILED(g_pDevice->Reset(&g_params)))
        return false;

    if (FAILED(g_pParallaxNormalMappingEffect->OnResetDevice()))
        return false;

    if (FAILED(g_pPerPixelLightingEffect->OnResetDevice()))
        return false;

    if (FAILED(g_pFont->OnResetDevice()))
        return false;

    return true;
}

void SetProcessorAffinity()
{
    // Assign the current thread to one processor. This ensures that timing
    // code runs on only one processor, and will not suffer any ill effects
    // from power management.
    //
    // Based on the DXUTSetProcessorAffinity() function in the DXUT framework.

    DWORD_PTR dwProcessAffinityMask = 0;
    DWORD_PTR dwSystemAffinityMask = 0;
    HANDLE hCurrentProcess = GetCurrentProcess();

    if (!GetProcessAffinityMask(hCurrentProcess, &dwProcessAffinityMask, &dwSystemAffinityMask))
        return;

    if (dwProcessAffinityMask)
    {
        // Find the lowest processor that our process is allowed to run against.

        DWORD_PTR dwAffinityMask = (dwProcessAffinityMask & ((~dwProcessAffinityMask) + 1));

        // Set this as the processor that our thread must always run against.
        // This must be a subset of the process affinity mask.

        HANDLE hCurrentThread = GetCurrentThread();

        if (hCurrentThread != INVALID_HANDLE_VALUE)
        {
            SetThreadAffinityMask(hCurrentThread, dwAffinityMask);
            CloseHandle(hCurrentThread);
        }
    }

    CloseHandle(hCurrentProcess);
}

void ToggleFullScreen()
{
    static DWORD savedExStyle;
    static DWORD savedStyle;
    static RECT rcSaved;

    g_isFullScreen = !g_isFullScreen;

    if (g_isFullScreen)
    {
        // Moving to full screen mode.

        savedExStyle = GetWindowLong(g_hWnd, GWL_EXSTYLE);
        savedStyle = GetWindowLong(g_hWnd, GWL_STYLE);
        GetWindowRect(g_hWnd, &rcSaved);

        SetWindowLong(g_hWnd, GWL_EXSTYLE, 0);
        SetWindowLong(g_hWnd, GWL_STYLE, WS_POPUP);
        SetWindowPos(g_hWnd, HWND_TOPMOST, 0, 0, 0, 0,
            SWP_NOMOVE | SWP_NOSIZE | SWP_FRAMECHANGED | SWP_SHOWWINDOW);

        g_windowWidth = GetSystemMetrics(SM_CXSCREEN);
        g_windowHeight = GetSystemMetrics(SM_CYSCREEN);

        SetWindowPos(g_hWnd, HWND_TOPMOST, 0, 0,
            g_windowWidth, g_windowHeight, SWP_SHOWWINDOW);

        // Update presentation parameters.

        g_params.Windowed = FALSE;
        g_params.BackBufferWidth = g_windowWidth;
        g_params.BackBufferHeight = g_windowHeight;

        if (g_enableVerticalSync)
        {
            g_params.FullScreen_RefreshRateInHz = D3DPRESENT_INTERVAL_DEFAULT;
            g_params.PresentationInterval = D3DPRESENT_INTERVAL_DEFAULT;
        }
        else
        {
            g_params.FullScreen_RefreshRateInHz = D3DPRESENT_INTERVAL_IMMEDIATE;
            g_params.PresentationInterval = D3DPRESENT_INTERVAL_IMMEDIATE;
        }
    }
    else
    {
        // Moving back to windowed mode.

        SetWindowLong(g_hWnd, GWL_EXSTYLE, savedExStyle);
        SetWindowLong(g_hWnd, GWL_STYLE, savedStyle);
        SetWindowPos(g_hWnd, HWND_NOTOPMOST, 0, 0, 0, 0,
            SWP_NOMOVE | SWP_NOSIZE | SWP_FRAMECHANGED | SWP_SHOWWINDOW);

        g_windowWidth = rcSaved.right - rcSaved.left;
        g_windowHeight = rcSaved.bottom - rcSaved.top;

        SetWindowPos(g_hWnd, HWND_NOTOPMOST, rcSaved.left, rcSaved.top,
            g_windowWidth, g_windowHeight, SWP_SHOWWINDOW);

        // Update presentation parameters.

        g_params.Windowed = TRUE;
        g_params.BackBufferWidth = g_windowWidth;
        g_params.BackBufferHeight = g_windowHeight;
        g_params.FullScreen_RefreshRateInHz = 0;

        if (g_enableVerticalSync)
            g_params.PresentationInterval = D3DPRESENT_INTERVAL_DEFAULT;
        else
            g_params.PresentationInterval = D3DPRESENT_INTERVAL_IMMEDIATE;
    }

    ResetDevice();

    // Viewport has changed in size. Rebuild the camera's projection matrix.
    g_camera.perspective(CAMERA_FOVX,
        static_cast<float>(g_windowWidth) / static_cast<float>(g_windowHeight),
        CAMERA_ZNEAR, CAMERA_ZFAR);
}

void UpdateCamera(float elapsedTimeSec)
{
    float dx = 0.0f;
    float dy = 0.0f;
    float dz = 0.0f;
    D3DXVECTOR3 direction;
    Mouse &mouse = Mouse::instance();

    GetMovementDirection(direction);

    switch (g_camera.getBehavior())
    {
    default:
        break;

    case Camera::CAMERA_BEHAVIOR_FIRST_PERSON:
    case Camera::CAMERA_BEHAVIOR_SPECTATOR:
        if (g_enableVerticalSync)
        {
            dx = mouse.xPosRelative() * elapsedTimeSec * MOUSE_SPEED_X;
            dy = mouse.yPosRelative() * elapsedTimeSec * MOUSE_SPEED_Y;
        }
        else
        {
            dx = mouse.xPosRelative();
            dy = mouse.yPosRelative();
        }

        g_camera.rotateSmoothly(dx, dy, 0.0f);
        g_camera.updatePosition(direction, elapsedTimeSec);
        break;

    case Camera::CAMERA_BEHAVIOR_FLIGHT:
        if (g_enableVerticalSync)
        {
            dy = -mouse.yPosRelative() * elapsedTimeSec * MOUSE_SPEED_Y;
            dz = mouse.xPosRelative() * elapsedTimeSec * MOUSE_SPEED_X;
        }
        else
        {
            dy = -mouse.yPosRelative();
            dz = mouse.xPosRelative();
        }

        g_camera.rotateSmoothly(0.0f, dy, dz);

        if ((dx = direction.x * CAMERA_SPEED_FLIGHT_YAW * elapsedTimeSec) != 0.0f)
            g_camera.rotate(dx, 0.0f, 0.0f);

        direction.x = 0.0f; // ignore yaw motion when updating camera's velocity
        g_camera.updatePosition(direction, elapsedTimeSec);
        break;

    case Camera::CAMERA_BEHAVIOR_ORBIT:
        if (g_enableVerticalSync)
        {
            dx = -mouse.xPosRelative() * elapsedTimeSec * MOUSE_SPEED_X;
            dy = -mouse.yPosRelative() * elapsedTimeSec * MOUSE_SPEED_Y;
        }
        else
        {
            dx = -mouse.xPosRelative();
            dy = -mouse.yPosRelative();
        }

        g_camera.rotateSmoothly(dx, dy, 0.0f);

        if (!g_camera.preferTargetYAxisOrbiting())
        {
            if ((dz = direction.x * CAMERA_SPEED_ORBIT_ROLL * elapsedTimeSec) != 0.0f)
                g_camera.rotate(0.0f, 0.0f, dz);
        }

        if ((dz = -mouse.wheelPos()) != 0.0f)
            g_camera.zoom(dz, g_camera.getOrbitMinZoom(), g_camera.getOrbitMaxZoom());

        break;
    }

    if (mouse.buttonPressed(Mouse::BUTTON_MIDDLE))
        g_camera.undoRoll();

    PerformCameraCollisionDetection();
}

void UpdateFrame(float elapsedTimeSec)
{
    Keyboard::instance().update();
    Mouse::instance().update();

    ProcessUserInput();

    UpdateFrameRate(elapsedTimeSec);
    UpdateCamera(elapsedTimeSec);
    UpdateParallaxNormalMappingEffectParameters();
    UpdatePerPixelLightingEffectParameters();
}

void UpdateFrameRate(float elapsedTimeSec)
{
    static float accumTimeSec = 0.0f;
    static int frames = 0;

    accumTimeSec += elapsedTimeSec;

    if (accumTimeSec > 1.0f)
    {
        g_framesPerSecond = frames;

        frames = 0;
        accumTimeSec = 0.0f;
    }
    else
    {
        ++frames;
    }
}

void UpdateParallaxNormalMappingEffectParameters()
{
    D3DXMATRIX identityMatrix;
    D3DXMATRIX viewProjMatrix = g_camera.getViewProjectionMatrix();

    D3DXMatrixIdentity(&identityMatrix);

    // The floor is centered about the world origin and doesn't move.
    // We can just use the identity matrix here for both the world and
    // normal matrices. The normal matrix is the transpose of the inverse of
    // the world matrix and is used to transform the mesh's normal vectors.
    // But since the floor isn't moving we can just use the identity matrix.

    g_pParallaxNormalMappingEffect->SetMatrix("worldMatrix", &identityMatrix);
    g_pParallaxNormalMappingEffect->SetMatrix("worldInverseTransposeMatrix", &identityMatrix);
    g_pParallaxNormalMappingEffect->SetMatrix("worldViewProjectionMatrix", &viewProjMatrix);

    g_pParallaxNormalMappingEffect->SetValue("cameraPos", &g_camera.getPosition(), sizeof(D3DXVECTOR3));
    g_pParallaxNormalMappingEffect->SetValue("globalAmbient", &g_globalAmbient, sizeof(g_globalAmbient));

    g_pParallaxNormalMappingEffect->SetValue("light.dir", g_light.dir, sizeof(g_light.dir));
    g_pParallaxNormalMappingEffect->SetValue("light.pos", g_light.pos, sizeof(g_light.pos));
    g_pParallaxNormalMappingEffect->SetValue("light.ambient", g_light.ambient, sizeof(g_light.ambient));
    g_pParallaxNormalMappingEffect->SetValue("light.diffuse", g_light.diffuse, sizeof(g_light.diffuse));
    g_pParallaxNormalMappingEffect->SetValue("light.specular", g_light.specular, sizeof(g_light.specular));
    g_pParallaxNormalMappingEffect->SetFloat("light.spotInnerCone", g_light.spotInnerCone);
    g_pParallaxNormalMappingEffect->SetFloat("light.spotOuterCone", g_light.spotOuterCone);
    g_pParallaxNormalMappingEffect->SetFloat("light.radius", g_light.radius);

    g_pParallaxNormalMappingEffect->SetValue("material.ambient", g_material.ambient, sizeof(g_material.ambient));
    g_pParallaxNormalMappingEffect->SetValue("material.diffuse", g_material.diffuse, sizeof(g_material.diffuse));
    g_pParallaxNormalMappingEffect->SetValue("material.emissive", g_material.emissive, sizeof(g_material.emissive));
    g_pParallaxNormalMappingEffect->SetValue("material.specular", g_material.specular, sizeof(g_material.specular));
    g_pParallaxNormalMappingEffect->SetFloat("material.shininess", g_material.shininess);

    if (g_disableColorMapTexture)
        g_pParallaxNormalMappingEffect->SetTexture("colorMapTexture", g_pNullTexture);
    else
        g_pParallaxNormalMappingEffect->SetTexture("colorMapTexture", g_pFloorColorMapTexture);

    g_pParallaxNormalMappingEffect->SetTexture("normalMapTexture", g_pFloorNormalMapTexture);

    if (g_enableParallax)
    {
        g_pParallaxNormalMappingEffect->SetValue("scaleBias", &g_parallaxScaleBias, sizeof(g_parallaxScaleBias));
        g_pParallaxNormalMappingEffect->SetTexture("heightMapTexture", g_pFloorHeightMapTexture);
    }
}

void UpdatePerPixelLightingEffectParameters()
{
    D3DXMATRIX worldMatrix;
    D3DXMATRIX worldViewProjMatrix;

    // Calculate world matrix.
    D3DXMatrixRotationQuaternion(&worldMatrix, &g_meshOrientation);
    worldMatrix(3, 0) = g_meshPosition.x;
    worldMatrix(3, 1) = g_meshPosition.y;
    worldMatrix(3, 2) = g_meshPosition.z;

    // Calculate combined world-view-projection matrix.
    worldViewProjMatrix = worldMatrix * g_camera.getViewProjectionMatrix();

    // Since the world matrix here is a rigid-body transformation (i.e., only
    // rotations and translations are used) we can simply use the world matrix
    // to transform the mesh's normals instead of computing the normal matrix.
    // The normal matrix is the transpose of the inverse of the world matrix.
    //
    // The reason why we don't have to compute the normal matrix is because:
    // (a) Rotations are usually orthogonal. This means that the inverse of an
    //     orthogonal rotation matrix is the same as taking the transpose of
    //     the orthogonal rotation matrix.
    // (b) Translations don't affect the direction of normals.

    g_pPerPixelLightingEffect->SetMatrix("worldMatrix", &worldMatrix);
    g_pPerPixelLightingEffect->SetMatrix("worldInverseTransposeMatrix", &worldMatrix);
    g_pPerPixelLightingEffect->SetMatrix("worldViewProjectionMatrix", &worldViewProjMatrix);
    
    g_pPerPixelLightingEffect->SetValue("cameraPos", &g_camera.getPosition(), sizeof(D3DXVECTOR3));
    g_pPerPixelLightingEffect->SetValue("globalAmbient", &g_globalAmbient, sizeof(g_globalAmbient));

    // Setup a directional light pointing in the direction of the camera.
    g_pPerPixelLightingEffect->SetValue("light.dir", &g_camera.getViewDirection(), sizeof(D3DXVECTOR3));
    g_pPerPixelLightingEffect->SetValue("light.ambient", g_light.ambient, sizeof(g_light.ambient));
    g_pPerPixelLightingEffect->SetValue("light.diffuse", g_light.diffuse, sizeof(g_light.diffuse));
    g_pPerPixelLightingEffect->SetValue("light.specular", g_light.specular, sizeof(g_light.specular));

    g_pPerPixelLightingEffect->SetTexture("colorMapTexture", g_pNullTexture);
}